import sys
import argparse
import logging
import importlib
import os
import asyncio

from .client import ChannelsIRCClient
from .multi import MultiConnectionClient

logger = logging.getLogger(__name__)


class CLI(object):
    """
    Primary command-line interface for irc -> channels bridge
    """
    description = "Django Channels <-> IRC Interface server"

    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description=self.description,
        )
        self.parser.add_argument(
            '-s',
            '--server',
            dest='server',
            help='IRC server to connect to. Can also be set from CHANNELS_IRC_SERVER env variable',
            default=os.environ.get('CHANNELS_IRC_SERVER', None),
        )
        self.parser.add_argument(
            '-p',
            '--port',
            type=int,
            help=(
                'Port number of IRC server to connect to. Can also be set from '
                'CHANNELS_IRC_PORT env variable. Default is 6667'
            ),
            default=os.environ.get('CHANNELS_IRC_PORT', 6667),
        )
        self.parser.add_argument(
            '-n',
            '--nickname',
            dest='nickname',
            help='Nickname on the IRC server',
            default=os.environ.get('CHANNELS_IRC_NICKNAME', None),
        )
        self.parser.add_argument(
            '--password',
            dest='password',
            help='Password on the IRC server',
            default=os.environ.get('CHANNELS_IRC_PASSWORD', None),
        )
        self.parser.add_argument(
            '--username',
            dest='username',
            help='Username on the IRC server. If none is provided, the value for `nickname` will be used',
            default=None,
        )
        self.parser.add_argument(
            '-r',
            '--realname',
            dest='realname',
            help='Real name on the IRC server.  If none is provided, the value for `nickname` will be used',
            default=None,
        )
        self.parser.add_argument(
            '-v',
            '--verbosity',
            type=int,
            help='How verbose to make the output. Default is 1',
            default=1,
        )
        self.parser.add_argument(
            '-a',
            '--application',
            dest='application',
            help='The application to dispatch to as path.to.module:instance.path',
            default=os.environ.get('CHANNELS_IRC_APPLICATION', None),
        )
        self.parser.add_argument(
            '--autoreconnect',
            dest='autoreconnect',
            action='store_true',
            help=(
                'Set autoreconnect to attempt to reconnect to the IRC server at regular intervals '
                'if disconnected'
            ),
        )
        self.parser.add_argument(
            '--reconnect-delay',
            dest='reconnect_delay',
            type=int,
            help='Time between reconnection attempts (in seconds). Default is 60',
            default=os.environ.get('CHANNELS_IRC_RECONNECT_DELAY', 60)
        )
        self.parser.add_argument(
            '--multi',
            dest='multi',
            action='store_true',
            help=(
                'Use to load the multi-connection server instead of the single connection server. '
                'Useful for things like chatbots that need to post from mutliple accounts.'
            ),
        )

    @classmethod
    def entrypoint(cls):
        """
        Many entrypoint for external starts
        """
        cls().run(sys.argv[1:])

    def run(self, args):
        """
        Mounts the IRC interface server based on the raw arguments passed in
        """
        args = self.parser.parse_args(args)

        # Set up logging
        logging.basicConfig(
            level={
                0: logging.WARN,
                1: logging.INFO,
                2: logging.DEBUG,
            }[args.verbosity],
            format="%(asctime)-15s %(levelname)-8s %(message)s",
        )

        # import the channel layer
        asgi_module, application_path = args.application.split(':', 1)
        application_module = importlib.import_module(asgi_module)
        for part in application_path.split('.'):
            application = getattr(application_module, part)

        # Parse bool flag values
        autoreconnect = os.environ.get('CHANNELS_IRC_RECONNECT', '') in ['true', 'True'] or args.autoreconnect
        multi = os.environ.get('CHANNELS_IRC_MULTI', '') in ['true', 'True'] or args.multi

        client_class = ChannelsIRCClient if not multi else MultiConnectionClient

        client = client_class(
            application,
            autoreconnect=autoreconnect,
            reconnect_delay=args.reconnect_delay,
        )

        if not multi:
            # If not a MultiConnectionClient, automatically connect to the IRC server
            if not all([args.application, args.server, args.nickname]):
                raise ValueError(
                    "--application, --server, and --nickname are required arguments. "
                    "Please add them via the command line or their respective env variables."
                )

            logger.info('Connecting to IRC Server {}:{}'.format(args.server, args.port))

            client.reactor.loop.run_until_complete(client.connect(
                args.server,
                args.port,
                args.nickname,
                password=args.password,
                username=args.username,
                ircname=args.realname,
            ))

        try:
            client.start()
        except KeyboardInterrupt:
            client.disconnect()

            loop = client.loop

            tasks = asyncio.gather(
                *asyncio.Task.all_tasks(loop=loop),
                loop=loop,
                return_exceptions=True
            )
            tasks.add_done_callback(lambda t: loop.stop())
            tasks.cancel()

            while not tasks.done() and not loop.is_closed():
                loop.run_forever()
        finally:
            loop.close()
            sys.exit(0)



from unittest.mock import patch, Mock

from ..consumers import AsyncIrcConsumer
from ..client import ChannelsIRCClient
from .utils import async_test, AsyncTestCase


class MockEvent(object):
    def __init__(self, **kwargs):
        self.source = kwargs.get('source', None)
        self.target = kwargs.get('target', None)
        self.arguments = kwargs.get('arguments', [''])
        self.type = kwargs.get('type', None)


class MockConnection(object):
    def __init__(self, **kwargs):
        self.server = kwargs.get('server', None)
        self.port = kwargs.get('port', None)


class ChannelsIRCClientTests(AsyncTestCase):
    @patch('irc.client.SimpleIRCClient.connect')
    def setUp(self, mock_connect):
        super().setUp()

        self.client = ChannelsIRCClient(AsyncIrcConsumer)
        self.loop.run_until_complete(self.client.connect('test.irc.server', 6667, 'advogg'))
        self.client.connection.send_raw = Mock()

        self.mock_connection = MockConnection(server='test.irc.server', port=6667)

    @async_test
    async def test_on_welcome(self):
        """
        According to the ASGI spec, when an IRC channel is joined, the client should
        return the a message on the `irc.connect` channel
        """
        mock_event = MockEvent(target='test-channel')
        self.client.on_welcome(self.mock_connection, mock_event)

        response = await self.client.application_queue.get()

        self.assertEqual(response, {
            'type': 'irc.receive',
            'channel': 'test-channel',
            'command': 'welcome',
        })

    @async_test
    async def test_on_join(self):
        """
        `join` doesn't have a specific in client handler, so it should send the generic 'irc.receive'
        with the relevant information
        """
        mock_event = MockEvent(target='#testchannel', type='join', args=[])
        self.client._dispatcher(self.mock_connection, mock_event)

        response = await self.client.application_queue.get()

        self.assertEqual(response, {
            'type': 'irc.receive',
            'command': 'join',
            'channel': '#testchannel',
            'body': [''],
        })

    @async_test
    async def test_handle_on_message(self):
        """
        `privmsg`, `pubmg`, and other messgate-related IRC events should be handled
        b the `_handle_on_message`, which should return the appropriately spec'ed data
        on the `irc.receive` channel
        """
        mock_event = MockEvent(target='#testchannel', source='testuser', arguments=['hello'])
        self.client._handle_on_message(self.mock_connection, mock_event)

        response = await self.client.application_queue.get()

        self.assertEqual(response, {
            'type': 'irc.receive',
            'channel': '#testchannel',
            'user': 'testuser',
            'command': 'message',
            'body': 'hello',
        })

    @async_test
    async def test_on_pubmsg_calls_handle_on_message(self):
        """
        `on_pubmsg` pass the event and connection to the `_on_handle_message` method
        """
        mock_event = MockEvent(target='#testchannel', source='testuser', arguments=['hello'])
        self.client.on_pubmsg(self.mock_connection, mock_event)

        response = await self.client.application_queue.get()

        self.assertEqual(response, {
            'type': 'irc.receive',
            'channel': '#testchannel',
            'user': 'testuser',
            'command': 'message',
            'body': 'hello',
        })

    @async_test
    async def test_on_privmsg_calls_handle_on_message(self):
        """
        `on_privmsg` pass the event and connection to the `_on_handle_message` method
        """
        mock_event = MockEvent(target='#testchannel', source='testuser', arguments=['hello'])
        self.client.on_privmsg(self.mock_connection, mock_event)

        response = await self.client.application_queue.get()

        self.assertEqual(response, {
            'type': 'irc.receive',
            'channel': '#testchannel',
            'user': 'testuser',
            'command': 'message',
            'body': 'hello',
        })

    @async_test
    async def test_handle_join_calls_send_raw(self):
        """
        `_handle_join` should take the Channels message and send the appropriate join
        to IRC
        """
        join_msg = {
            'type': 'irc.send',
            'command': 'join',
            'channel': '#advogg',
        }

        await self.client._handle_join(join_msg)

        self.client.connection.send_raw.assert_called_with('JOIN #advogg')

    @async_test
    async def test_handle_join_adds_missing_hashtag(self):
        """
        `_handle_join` should add # to the channel name if it's missing
        """
        join_msg = {
            'type': 'irc.send',
            'command': 'join',
            'channel': 'advogg',
        }

        await self.client._handle_join(join_msg)

        self.client.connection.send_raw.assert_called_with('JOIN #advogg')

    @async_test
    async def test_handle_join_does_nothing_if_no_channel_is_specified(self):
        """
        `_handle_join` should not be called if no channel is given
        """
        join_msg = {
            'type': 'irc.send',
            'command': 'join',
        }

        await self.client._handle_join(join_msg)

        self.client.connection.send_raw.assert_not_called()

    @async_test
    async def test_handle_message_calls_send_raw(self):
        """
        `_handle_message` should call `send_raw` with the appropriate PRIVMSG text
        """
        privmsg = {
            'type': 'irc.send',
            'command': 'message',
            'channel': 'advogg',
            'body': 'Hello World!',
        }

        await self.client._handle_message(privmsg)

        self.client.connection.send_raw.assert_called_with('PRIVMSG #advogg :Hello World!')

    @async_test
    async def test_handle_part_calls_send_raw(self):
        """
        `_handle_part` should call `send_raw` with the appropriate PART message
        """
        part_msg = {
            'type': 'irc.send',
            'command': 'part',
            'channel': 'advogg',
        }

        await self.client._handle_part(part_msg)

        self.client.connection.send_raw.assert_called_with('PART #advogg')



from channels.consumer import AsyncConsumer
from channels.exceptions import InvalidChannelLayerError, StopConsumer


class AsyncIrcConsumer(AsyncConsumer):
    """
    Base IRC consumer; Implements basic hooks for interfacing with the IRC Interface Server
    """
    groups = []

    async def on_welcome(self, channel, user=None, body=None):
        """
        Called when the IRC Interface Server connects to the IRC Server
        and receives the "welcome" command from IRC
        """
        try:
            for group in self.groups:
                await self.channel_layer.group_add(group, self.channel_name)
        except AttributeError:
            raise InvalidChannelLayerError("BACKEND is unconfigured or doesn't support groups")
        await self.welcome(channel)

    async def welcome(self, channel):
        """
        Hook for any action(s) to be run on connecting to the IRC Server
        """
        pass

    async def irc_on_disconnect(self, message):
        """
        Called when the connection to the IRC Server is closed
        """
        try:
            for group in self.groups:
                await self.channel_layer.group_discard(group, self.channel_name)
        except AttributeError:
            raise InvalidChannelLayerError("BACKEND is unconfigured or doesn't support groups")

        await self.on_disconnect(message['server'][0], message['server'][1])
        raise StopConsumer()

    async def on_disconnect(self, server, port):
        """
        Hook for any action(s) to be run on disconnecting from the IRC Server
        """
        pass

    async def irc_receive(self, message):
        """
        Parses incoming messages and routes them to the appropriate handler, depending on the
        incoming action type
        """
        command_type = message.get('command', None)

        if command_type is None:
            raise ValueError('An `irc.receive` message must specify a `command` key')

        handler = getattr(self, 'on_{}'.format(command_type), None)

        if handler is not None:
            await handler(
                channel=message.get('channel', None),
                user=message.get('user', None),
                body=message.get('body', None),
            )

    async def send_message(self, channel, text):
        """
        Sends a PRIVMSG to the IRC Server
        """
        await self.send_command('message', channel=channel, body=text)

    async def send_command(self, command, channel=None, body=None):
        """
        Sends a command to the IRC Server.  Message should be of the format:
            {
                'type': 'irc.command',
                'command': '<IRC_COMMAND>',
                'channel': '<IRC_CHANNEL>',  # Optional, depending on command
                'body': '<COMMAND_TEXT>',  # Optional, depending on command
            }
        """
        await self.send({
            'type': 'irc.send',
            'command': command,
            'channel': channel,
            'body': body,
        })


class MultiIrcConsumer(AsyncConsumer):
    """
    Consumer for managing multiple IRC connections.  Used with the `MultiConnectionClient`
    """
    async def irc_multi_init(self, message):
        """
        Called when the consumer is initially loaded.
        """
        await self.on_init()

    async def on_init(self):
        """
        Hook for executing commands on loading the consumer
        """
        pass

    async def send_connect(self, server, port, nickname, **kwargs):
        """
        Creates a new connection, if no connection to that server/nickname
        combination exists.
        """
        await self.send({
            'type': 'irc.multi.connect',
            'server': server,
            'port': port,
            'nickname': nickname,
            **kwargs
        })

    async def send_disconnect(self, server, nickname):
        """
        Disconnects a connnect and removes it from the stored connections
        """
        await self.send({
            'type': 'irc.multi.disconnect',
            'server': server,
            'nickname': nickname,
        })



import asyncio

from .client import ChannelsIRCClient
from .server import BaseServer


class MultiConnectionClient(BaseServer):
    def __init__(self, application, autoreconnect=False, reconnect_delay=60, loop=None):
        self.application = application
        self.autoreconnect = autoreconnect
        self.reconnect_delay = reconnect_delay

        self.loop = loop if loop is not None else asyncio.get_event_loop()

        # dictionary of 'SERVER:NICKNAME': ChannelsIRCClient
        self.connections = {}

        self.create_application(
            scope={'type': 'irc.multi'}, from_consumer=self.from_consumer
        )
        self.send_init()
        self.loop.call_later(1, self.futures_checker)

    def start(self):
        self.loop.run_forever()

    def send_init(self):
        """
        Sent when the applcation is initialized.
        Used as a hook for creating any automatic connections
        """
        self._send_application_msg({
            'type': 'irc.multi.init',
        })

    async def from_consumer(self, message):
        """
        Receives message from channels from the consumer.
        """
        if 'type' not in message:
            raise ValueError('Message has no type defined')

        elif message['type'] == 'irc.multi.connect':
            await self.create_connection(**message)

        elif message['type'] == 'irc.multi.disconnect':
            await self.remove_connection(message['server'], message['nickname'])

        else:
            raise ValueError("Cannot handle message type %s!" % message["type"])

    def get_connection_key(self, server, nickname):
        """
        cretes the key for a connection in `self.connections`
        """
        return '{}:{}'.format(server, nickname)

    async def create_connection(self, server, port, nickname, **kwargs):
        """
        Create an instance of `ChannelsIRCClient`, and store it in self.connections
        under the key of `SERVER:NICKNAME`.  To ensure idempotency, it first checks
        whether the connection exists and is already connected before attempt to
        start a new connection
        """
        key = self.get_connection_key(server, nickname)
        connection = self.connections.get(key, None)

        if connection is None or not connection.connected:
            client = ChannelsIRCClient(
                self.application, autoreconnect=self.autoreconnect,
                reconnect_delay=self.reconnect_delay, loop=self.loop
            )

            kwargs.pop('type')
            await client.connect(server, port, nickname, **kwargs)

            self.connections[key] = client

    async def remove_connection(self, server, nickname):
        """
        Shuts down the connection, clean up tasks, remove from self.connections
        """
        key = self.get_connection_key(server, nickname)

        connection = self.connections[key]

        if connection is not None:
            connection.disconnect()

            if not connection.application_instance.done():
                connection.application_instance.cancel()
                await asyncio.wait([connection.application_instance])

            self.connections.pop(key, None)
            connection = None

    def disconnect(self):
        """
        Disconnect from all active connections
        """
        for connection in self.connections.values():
            connection.disconnect()



import asyncio
import traceback
import logging

logger = logging.getLogger(__name__)


class BaseServer:
    """
    Base class for common interface server functions, including
    creating an application, sending to the consumer, and application_checking/
    error handling
    """

    def _send_application_msg(self, msg):
        """
        sends a msg (serializable dict) to the appropriate Django channel
        """
        return self.application_queue.put_nowait(msg)

    def noop_from_consumer(self, msg):
        """
        empty default for receiving from consumer
        """

    def create_application(self, scope={}, from_consumer=noop_from_consumer):
        """
        Handles creating the ASGI application and instatiating the
        send Queue
        """
        application_instance = self.application(scope=scope)
        self.application_queue = asyncio.Queue()
        self.application_instance = asyncio.ensure_future(application_instance(
            receive=self.application_queue.get,
            send=from_consumer),
            loop=self.loop
        )

    def futures_checker(self):
        """
        Looks for exeptions raised in the application or irc loops
        """
        if getattr(self, 'application_instance', False) and self.application_instance.done():
            try:
                exception = self.application_instance.exception()
            except asyncio.CancelledError:
                # Future cancellation. We can ignore this.
                pass
            else:
                if exception:
                    exception_output = "{}\n{}{}".format(
                        exception,
                        "".join(traceback.format_tb(
                            exception.__traceback__,
                        )),
                        "  {}".format(exception),
                    )
                    logger.error(
                        "Exception inside application: %s",
                        exception_output,
                    )
                    self.disconnect()

                del self.application_instance
                self.application_instance = None

        self.loop.call_later(1, self.futures_checker)



SECRET_KEY = "it's-a-secret-to-everyone"

INSTALLED_APPS = [
    'channels',
    'channels_irc',
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels.layers.InMemoryChannelLayer",
    },
}



from unittest.mock import patch

from channels.testing import ApplicationCommunicator

from .utils import AsyncTestCase, AsyncMock, async_test
from ..consumers import AsyncIrcConsumer


class AsyncIrcConsumerTests(AsyncTestCase):
    @patch('channels_irc.consumers.AsyncIrcConsumer.welcome', new_callable=AsyncMock)
    @async_test
    async def test_on_welcome(self, mock_welcome):
        """
        `irc.receive` called with a `welcome` command should call the
        `on_welcome`
        """
        communicator = ApplicationCommunicator(AsyncIrcConsumer, {'type': 'irc'})

        await communicator.send_input({
            'type': 'irc.receive',
            'command': 'welcome',
            'channel': 'test_channel',
            'user': 'my_nick',
            'body': None,
        })

        await communicator.wait(timeout=.2)
        self.assertEqual(mock_welcome.call_count, 1)

    @async_test
    async def test_send_command(self):
        """
        `send_command` should format the correct message and return it to the
        server
        """
        communicator = ApplicationCommunicator(AsyncIrcConsumer, {'type': 'irc'})

        # Give the loop a beat to initialize the instance
        await communicator.receive_nothing(timeout=.2)

        await communicator.instance.send_command('join', channel='my_channel')

        event = await communicator.receive_output(timeout=1)
        self.assertEqual(event, {
            'type': 'irc.send',
            'command': 'join',
            'channel': 'my_channel',
            'body': None,
        })

    @async_test
    async def test_send_message(self):
        """
        `send_message` should format the correct message and return it to the
        server
        """
        communicator = ApplicationCommunicator(AsyncIrcConsumer, {'type': 'irc'})

        # Give the loop a beat to initialize the instance
        await communicator.receive_nothing(timeout=.2)

        await communicator.instance.send_message('my_channel', 'Hello IRC!')

        event = await communicator.receive_output(timeout=1)
        self.assertEqual(event, {
            'type': 'irc.send',
            'command': 'message',
            'channel': 'my_channel',
            'body': 'Hello IRC!',
        })



import asyncio
from unittest.mock import Mock

from django.test import TestCase


def async_test(coro):
    """
    Async test wrapper courtesy of: https://stackoverflow.com/a/23036785/3362468
    """
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(coro(*args, **kwargs))
    return wrapper


def AsyncMock():
    """
    Substitute Mock for mocking/patching coroutines.  Courtesy of:
    https://stackoverflow.com/a/32505333/3362468
    """
    coro = Mock(name="CoroutineResult")
    corofunc = Mock(name="CoroutineFunction",
                    side_effect=asyncio.coroutine(coro))
    corofunc.coro = coro
    return corofunc


class AsyncTestCase(TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        tasks = asyncio.gather(
            *asyncio.Task.all_tasks(loop=self.loop),
            loop=self.loop,
            return_exceptions=True
        )
        tasks.add_done_callback(lambda t: self.loop.stop())
        tasks.cancel()

        while not tasks.done() and not self.loop.is_closed():
            self.loop.run_forever()

        self.loop.close()



import logging
import asyncio
from socket import gaierror

from irc.client_aio import AioSimpleIRCClient

from .server import BaseServer

logger = logging.getLogger(__name__)


class ChannelsIRCClient(AioSimpleIRCClient, BaseServer):
    def __init__(self, application, autoreconnect=False, reconnect_delay=60, loop=None):
        self.application = application
        self.autoreconnect = autoreconnect
        self.reconnect_delay = reconnect_delay

        self.reactor = self.reactor_class(loop=loop)
        self.connection = self.reactor.server()
        self.loop = self.reactor.loop

        self.reactor.add_global_handler("all_events", self._dispatcher, -10)
        self.loop.call_later(1, self.futures_checker)

    @property
    def connected(self):
        return getattr(self.connection, 'connected', False)

    def _dispatcher(self, connection, event):
        method = getattr(self, "on_" + event.type, None)

        if method is not None:
            method(connection, event)
        else:
            self._send_application_msg({
                'type': 'irc.receive',
                'command': event.type,
                'body': event.arguments,
                'channel': event.target,
            })

    def on_welcome(self, connection, event):
        """
        Sends `irc.receive` with welcome info
        """
        logger.info('Connected to IRC Server {}:{}'.format(connection.server, connection.port))

        msg = {
            'type': 'irc.receive',
            'command': 'welcome',
            'channel': event.target,
        }
        self._send_application_msg(msg)

    def on_privmsg(self, connection, event):
        """
        Sends message to `irc.receive` with incoming info
        """
        self._handle_on_message(connection, event)

    def on_pubmsg(self, connection, event):
        """
        Sends message to `irc.receive` with incoming info
        """
        self._handle_on_message(connection, event)

    def on_disconnect(self, connection, event):
        """
        Sends message type `irc.disconnected` with disconnected server info
        """
        msg = {
            'type': 'irc.on.disconnect',
            'server': [connection.server, connection.port],
        }
        self._send_application_msg(msg)

    def _handle_on_message(self, connection, event):
        msg = {
            'type': 'irc.receive',
            'command': 'message',
            'user': event.source,
            'channel': event.target,
            'body': event.arguments[0],
        }
        self._send_application_msg(msg)

    def disconnect(self, message=""):
        """
        Disconnects from the current IRC connection
        """
        logger.info("Disconnecting from {}:{}...".format(
            self.connection.server, self.connection.port
        ))
        self.connection.disconnect(message=message)

    async def connect(
        self,  server, port, nickname, is_reconnect=False, *args, **kwargs
    ):
        """
        Instantiates the connection to the server.  Also creates the requisite
        application instance
        """
        scope = {
            'type': 'irc',
            'server': server,
            'port': port,
            'nickname': nickname,
        }
        self.create_application(scope=scope, from_consumer=self.from_consumer)

        try:
            await self.connection.connect(
                server, port, nickname, *args, **kwargs
            )
        except gaierror:
            logger.debug('Connection attempt to {} with user {} failed '.format(
              server, nickname
            ))

        if self.autoreconnect and not is_reconnect:
            self.loop.call_later(self.reconnect_delay, self.reconnect_checker)

    async def from_consumer(self, message):
        """
        Receives message from channels from the consumer.  Message should have the format:
            {
                'type': 'irc.send',
                'command': VALID_COMMAND_TYPE
            }
        """
        if "type" not in message:
            raise ValueError("Message has no type defined")

        elif message['type'] == 'irc.send':
            command_type = message.get('command', '').lower()

            handler = getattr(self, '_handle_{}'.format(command_type))
            await handler(message)

        else:
            raise ValueError("Cannot handle message type %s!" % message["type"])

    async def _handle_status(self, msg):
        """
        Gets the current status of the IRC Interface server, returns that information to
        channels
        """
        self._send_application_msg({
            'type': 'irc.receive',
            'command': 'status',
            'body': {
                'connected': self.connected,
            },
        })

    async def _handle_join(self, msg):
        """
        Makes join call as requested from channels.  Channels msg should be in format:
            {
                'type': 'irc.send',
                'command': 'join',
                'channel': <CHANNEL_NAME>,
            }
        """
        channel = msg.get('channel', '')

        if channel:
            self.connection.join(self.format_channel_name(channel))

    async def _handle_message(self, msg):
        """
        Posts a message to the passed channel from Django channels. Channel msg should be in format:
            {
                'type': 'irc.send',
                'command': 'message',
                'channel': <CHANNEL_NAME>,
                'body': <MESSAGE_TEXT>,
            }
        """
        channel = msg.get('channel', '')

        message = msg.get('body', '')
        self.connection.privmsg(self.format_channel_name(channel), message)

    async def _handle_names(self, msg):
        """
        Posts a NAMES request to the passed channel from Django channels.
        Channel msg should be in the format:
            {
            'type': 'irc.send',
            'command': 'names',
            'channel': <CHANNEL_NAME>,
            }
        """
        channel = msg.get('channel', '')

        if channel:
            self.connection.names(self.format_channel_name(channel))

    async def _handle_part(self, msg):
        """
        Handles PART commands
        """
        channel = msg.get('channel', '')

        if channel:
            self.connection.part(self.format_channel_name(channel))

    async def _handle_disconnect(self, msg):
        """
        a DISCONNECT command should disconnect from the IRC server. Channel msg should be in
        the format:
            {
            'type': 'irc.send',
            'command': 'disconnect',
            'body': <MESSAGE_TEXT>,  # optional disconnect message
            }
        """
        message = msg.get('body', '')
        self.disconnect(message=message)

    def reconnect_checker(self):
        """
        Checks at a regular interval as to whether the connection to IRC has been
        disconnected, and re-starts the connection if it has
        """
        if not self.connected and self.autoreconnect:
            logger.info('Attempting to reconnect to {}:{}'.format(
                self.connection.server, self.connection.port
            ))
            asyncio.ensure_future(
                self.connect(
                    self.connection.server,
                    self.connection.port,
                    self.connection.nickname,
                    password=self.connection.password,
                    username=self.connection.username,
                    ircname=self.connection.ircname,
                    is_reconnect=True,
                ),
                loop=self.loop
            )

        self.loop.call_later(self.reconnect_delay, self.reconnect_checker)

    def start(self):
        super().start()

    def format_channel_name(self, name):
        """
        Adds missing `#` if necessary
        """
        return name if name[0] == '#' else '#{}'.format(name)



import os
from setuptools import setup, find_packages

from channels_irc import __version__

readme = os.path.join(os.path.dirname(__file__), 'PYPI_README.rst')

setup(
    name='channels_irc',
    version=__version__,
    url='https://github.com/AdvocatesInc/django-channels-irc',
    author='Advocates, Inc',
    author_email='admin@adv.gg',
    description='Interface server connecting Django\'s channels and IRC',
    long_description=open(readme).read(),
    license='Proprietary and confidential',
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'irc>=16.4',
        'asgiref>=2.3.0',
        'channels>=2.0.2',
    ],
    entry_points={'console_scripts': [
        'channels-irc = channels_irc.cli:CLI.entrypoint'
    ]},
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import asyncio
from unittest.mock import patch, MagicMock

from ..multi import MultiConnectionClient
from ..consumers import MultiIrcConsumer
from ..client import ChannelsIRCClient
from .utils import async_test, AsyncMock, AsyncTestCase


class MultiConnectionClientTests(AsyncTestCase):
    def make_fake_client(self):
        client = ChannelsIRCClient(MagicMock)
        client.application_instance = asyncio.Future()
        return client

    @async_test
    async def test_send_init(self):
        """
        On creating a new instance of `MultiConnectionClient` it should
        send the `irc.multi.init` message
        """
        client = MultiConnectionClient(MultiIrcConsumer)

        response = await client.application_queue.get()
        self.assertEqual(response, {'type': 'irc.multi.init'})

    @patch('channels_irc.multi.ChannelsIRCClient.connect', new_callable=AsyncMock)
    @async_test
    async def test_creating_new_connection(self, mock_connect):
        """
        Sending a `irc.multi.connect` message to the `MultiConnectionClient`
        should spin up a new connection with the appropriate server, nickname, etc,
        and add it to the connections dict
        """
        msg = {
            'type': 'irc.multi.connect',
            'server': 'my.test.server',
            'port': 6667,
            'nickname': 'my_nick',
        }

        client = MultiConnectionClient(MultiIrcConsumer)

        self.assertEqual(len(client.connections), 0)
        await client.from_consumer(msg)

        self.assertEqual(len(client.connections), 1)
        self.assertIn('my.test.server:my_nick', client.connections)

        mock_connect.assert_called_with(
            'my.test.server', 6667, 'my_nick',
        )

    @patch('channels_irc.multi.ChannelsIRCClient.disconnect', new_callable=AsyncMock)
    @async_test
    async def test_disconnecting(self, mock_disconnect):
        """
        Sending a `irc.multi.disconnect` message to the `MultiConnectionClient`
        should disconnect the connection and remove it from the connections dict
        """
        msg = {
            'type': 'irc.multi.disconnect',
            'server': 'my.test.server',
            'port': 6667,
            'nickname': 'my_nick',
        }

        client = MultiConnectionClient(MultiIrcConsumer)

        client.connections = {
            'my.test.server:my_nick': self.make_fake_client(),
            'my.test.server:other_nick': self.make_fake_client(),
        }
        self.assertEqual(len(client.connections), 2)
        await client.from_consumer(msg)

        self.assertEqual(len(client.connections), 1)
        self.assertNotIn('my.test.server:my_nick', client.connections)

        self.assertEqual(mock_disconnect.call_count, 1)
